! CGV/skel/cgv_04.F90

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH

! >>>>>>>>>>>>>>>>>>>>>> begin of task 00 <<<<<<<<<<<<<<<<<<<<<<

PROGRAM cgv_fortran

#ifndef serial
  INCLUDE 'mpif.h'
#endif

! ****************** declarations and global data ***************

! --- options: ---
INTEGER :: m,n         ! size of the physical-heat matrix without borders  
                     ! m=vertical, n=horizontal 
INTEGER :: iter_max    ! maximal number of iterations in CG
REAL :: epsilon        ! Limit for sqr( || residual vector || ) 
INTEGER :: print_level ! print_level = 1: Only result-error and partial result matrix 
                                 ! = 2: and residual norm after each iteration 
                                 ! = 3: and result physical-heat matrix 
                                 ! = 4: and all vector and halo data in first iteration 
                                 ! = 5: and all vector and halo data in _all_ iterations
INTEGER :: decomp_dims      ! dimensins of the domain decomposition 
INTEGER :: m_procs, n_procs ! number of processes in both directions of the 2-dim domain decomposition 
INTEGER :: mprt, nprt       ! number of rows and columns when printing the partial result matrix 

! global structures and data: 

#ifndef serial
  INTEGER :: ierror, comm_cart
  DOUBLE PRECISION :: commtime_start=0, commtime_end=0, commtime_sum=0
#endif
INTEGER :: numprocs, my_rank
REAL :: flops

! size of the physical area this process is about to calculate
! start_i   : start row
! end1_i    : end row + 1
! start_j   : start column
! end1_j    : end column + 1
! rowsize   : number of data entries in a row of the local chunk
!             (end1_j - start_j)
! colsize   : number of data entries in a column of the local chunk
!             (end1_i - start_i)
! chunksize : number of data entries in the local chunk PLUS halo data
!             ((rowsize + 2) * (colsize + 2))
INTEGER :: start_i, end1_i, start_j, end1_j, rowsize, colsize, chunksize

! SIDX (Serial InDeX): computes the serial index of a two-dimensional array 
!   input: i = 0...(colsize + 1) and j = 0...(rowsize + 1) 
!   output: SIDX in 0..((colsize + 2) * (rowsize + 2) - 1)
#define SIDX(i, j) ((rowsize + 2) * (i) + (j))

! --- data structures ---

! halo structure
TYPE halo_struct
    REAL, DIMENSION(:), POINTER :: north
    INTEGER                     :: north_rank
    INTEGER                     :: north_size
    REAL, DIMENSION(:), POINTER :: east
    INTEGER                     :: east_rank
    INTEGER                     :: east_size
    REAL, DIMENSION(:), POINTER :: south
    INTEGER                     :: south_rank
    INTEGER                     :: south_size
    REAL, DIMENSION(:), POINTER :: west
    INTEGER                     :: west_rank
    INTEGER                     :: west_size
END TYPE halo_struct
TYPE(halo_struct) :: halo

! needed global vectors
REAL, DIMENSION(:), ALLOCATABLE :: x, b
! needed send and receive scratch buffers
REAL, DIMENSION(:), ALLOCATABLE :: recv_scratch_buff_east, recv_scratch_buff_west, &
                                 & recv_scratch_buff_north, recv_scratch_buff_south

! >>>>>>>>>>>>>>>>>>>>>> -end- of task 00 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 08 <<<<<<<<<<<<<<<<<<<<<<

!************************ main program *************************

    INTEGER                         :: num_iter
    REAL                            :: norm_r, norm_err, maxnorm_err
    REAL, DIMENSION(:), ALLOCATABLE :: u, e
    REAL                            :: time_start_cpu, time_end_cpu, time_elapsed_cpu
#ifndef serial
    DOUBLE PRECISION :: time_start=0, time_end=0
    INTEGER :: dims(0:1)
    INTEGER :: percentage
#endif
    REAL :: mflops = 0, global_mflops = 0, time_elapsed, global_mflops_cpu

#ifdef serial
    numprocs = 1
    my_rank = 0
#endif
#ifndef serial
    CALL MPI_INIT(ierror)
    CALL MPI_COMM_SIZE(MPI_COMM_WORLD, numprocs, ierror)
    CALL MPI_COMM_RANK(MPI_COMM_WORLD, my_rank, ierror)
#endif

    flops = 0
    m = 4
    n = 4
    iter_max = 500
    epsilon = 1e-6
    print_level = 1
    decomp_dims = 1
    mprt = 11
    nprt = 11
    m_procs=1
    n_procs=1
#ifndef serial
    dims(0) = 0
    dims(1) = 0
    CALL MPI_DIMS_CREATE(numprocs, 2, dims, ierror)
    m_procs = dims(0)
    n_procs = dims(1)
#endif

    ! Parse command line options
    CALL read_options()

    ! Init the domain decomposition
    CALL Init_Decomp()
    IF (print_level >= 1) THEN
        CALL print_decomp()
    END IF

    ! reserve chunk of vector b and assign border data
    ALLOCATE(b(0:chunksize-1))
    b = 0
    CALL Init_Boundary_Vec(b)

    IF (print_level>=4) THEN
        CALL print_vec('b',b)
    END IF

    ! x is initialized with zeros
    ALLOCATE(x(0:chunksize-1))
    x = 0

#ifndef serial
    time_start = MPI_WTIME()
#endif
    CALL CPU_TIME(time_start_cpu)

    CALL CG_solve(num_iter, norm_r)

#ifndef serial
    time_end = MPI_WTIME()
    time_elapsed = time_end - time_start
    mflops = flops / (1000000 * time_elapsed)
    CALL MPI_REDUCE(mflops, global_mflops, 1, MPI_REAL, MPI_SUM, 0, comm_cart, ierror)
#endif
    CALL CPU_TIME(time_end_cpu)
    time_elapsed_cpu = time_end_cpu - time_start_cpu
    global_mflops_cpu = flops / (1000000 * time_elapsed_cpu)
#ifndef serial
    global_mflops_cpu = global_mflops
#endif

    ALLOCATE(u(0:chunksize-1))
    u = 0
    ALLOCATE(e(0:chunksize-1))
    e = 0
    CALL Init_Exact_Solution_vec(u)
    ! p = u - x
    CALL add_vec_vec(u, -1.0, x, e)
    norm_err = SQRT(sqr_norm_vec(e))
    maxnorm_err = max_norm_vec(e)

    IF (print_level>=3) THEN
        CALL print_vec2d('x',x,m,n)
    END IF

    IF (my_rank==0) THEN
        WRITE(*,'(A, I5, A, E8.2, A, E8.2)') 'main: ', num_iter, ' iterations, sqrt(eps) = ', SQRT(epsilon), &
                 & ',  normalized: ', SQRT(epsilon/(n*m))
        WRITE(*,'(A, E8.2, A, E8.2)') 'main:                 ||r||     = ', norm_r, ',  normalized: ', norm_r/SQRT(REAL (n*m))
        WRITE(*,'(A, E8.2, A, E8.2)') 'main:                 ||error|| = ', norm_err, ',  normalized: ', norm_err/SQRT(REAL (n*m))
        WRITE(*,'(A, E8.2, A, E8.2)') 'main:                 max(error)= ', maxnorm_err, ' = normalized: ', maxnorm_err
        WRITE(*,'(A, I0, A, I0, A, I0)') 'main:                 n*m = ', n, ' *', m, ' = ', n*m
        WRITE(*,'(A, E8.2, A)') 'main:                 ', time_elapsed_cpu, ' sec (exec. time CPU_TIME)'
#ifndef serial
        WRITE(*,'(A, E8.2, A)') 'main:                 ', time_elapsed, ' sec (exec. time MPI_WTIME)'
        WRITE(*,'(A, E8.2, A)') 'main:                 ', commtime_sum, ' sec (comm. time)'
        percentage = commtime_sum * 100.0 / time_elapsed
        WRITE(*,'(A, I2, A)') 'main: ratio (comm. time) / (exec. time): ', percentage, '%'
#endif
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (avg) CPU_TIME:  ', global_mflops_cpu / numprocs
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (all) CPU_TIME:  ', global_mflops_cpu
#ifndef serial
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (avg) MPI_WTIME:  ', global_mflops / numprocs
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (all) MPI_WTIME:  ', global_mflops
#endif
    END IF

    IF (print_level>=1) THEN
        CALL print_vec2d('x',x,mprt,nprt)
    END IF

#ifndef serial
    CALL MPI_FINALIZE(ierror)
#endif

! beginning of the subroutine part
CONTAINS

! ********************** memory management **********************

SUBROUTINE abrt
#ifndef serial
    CALL MPI_ABORT(MPI_COMM_WORLD, -1, ierror)
#endif
    STOP 
END SUBROUTINE abrt

! ********************** commandline options ********************

SUBROUTINE read_options()
    INTEGER :: optid = 1 ! start at first optional argument 
    INTEGER :: argc
    LOGICAL :: err_detected = .FALSE.
    CHARACTER(len=255) :: arg
    argc = iargc()
    IF ( my_rank == 0 ) THEN
        DO WHILE( optid <= argc )
            CALL getarg(optid, arg) 
            IF (arg == '-m') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) m
                IF (m<=0) THEN
                    err_detected = .TRUE.
                    EXIT 
                END IF
            ELSE IF (arg == '-n') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) n
                IF (n<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                ENDIF
            ELSE IF (arg == '-imax') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) iter_max
                IF (iter_max<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                ENDIF
            ELSE IF (arg == '-eps') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) epsilon
                IF (epsilon<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE IF (arg == '-prtlev') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) print_level
                IF (print_level<0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE IF (arg == '-twodims') THEN
                decomp_dims = 2
            ELSE IF (arg == '-mprocs') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) m_procs
                IF (m_procs<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE IF (arg == '-nprocs') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) n_procs
                IF (n_procs<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE
                err_detected = .TRUE.
                EXIT
            END IF
            optid = optid + 1
        END DO
        IF (err_detected) THEN
            WRITE(*,*) 'unknown command line option: ', arg
            WRITE(*,*) 'Synopsis: mpirun -np <number of processes> ./cg_all <options>'
            WRITE(*,*) 'with options [defaults]:'
            WRITE(*,*) '  -m <number of rows in physical area of the Laplace equation> [4]'
            WRITE(*,*) '  -n <number of columns in physical area of the Laplace equation [4]'
            WRITE(*,*) '  -imax <maximal number of iterations> [500]'
            WRITE(*,*) '  -eps  <epsion> [1e-6]'
            WRITE(*,*) '  -prtlev <print and debug level 0..5> [1]'
            WRITE(*,*) '  -twodims [only one-dimensional domain decomposition (DD)]'
            WRITE(*,*) '  -mprocs <horizontal number of processes in 2-dim DD'
            WRITE(*,*) '  -nprocs <vertical   number of processes in 2-dim DD'
            WRITE(*,*) ''
            CALL abrt
        END IF
    END IF
#ifndef serial
    CALL MPI_BCAST(m,          1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(n,          1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(iter_max,   1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(epsilon,    1, MPI_REAL,       0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(print_level,1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(decomp_dims,1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(m_procs,    1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(n_procs,    1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
#endif
END SUBROUTINE read_options

! >>>>>>>>>>>>>>>>>>>>>> -end- of task 08 <<<<<<<<<<<<<<<<<<<<<<
! >>>>>>>>>>>>>>>>>>>>>> begin of task 01 <<<<<<<<<<<<<<<<<<<<<<

! ********************** vector routines ************************

!-------------------------------------------
! add_vec_vec
!  IN  REAL(0:) v1      : vector 1
!  IN  REAL     lambda  : factor lambda
!  IN  REAL(0:) v2      : vector 2
!  OUT REAL(0:) v3      : vector 3
!
! RETURNS:
!   --
!
! TASK:
!   calculate:   v3 := v1 + lambda * v2
!
! GLOBAL DATA:
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT REAL    flops   : amount of FLOPS
!
! USED FUNCTIONS:
!   --
SUBROUTINE add_vec_vec(v1, lambda, v2, v3)
    REAL, DIMENSION(0:), INTENT(IN)  ::  v1
    REAL, INTENT(IN)                 ::  lambda
    REAL, DIMENSION(0:), INTENT(IN)  ::  v2
    REAL, DIMENSION(0:), INTENT(OUT) ::  v3
    INTEGER                          ::  i

    ! addition of two vectors, second vector is multiplied by a scalar
    DO i = (rowsize + 2), ((rowsize + 2) * (colsize + 1)) - 1
        v3(i) = v1(i) + lambda * v2(i);
    END DO

    ! for the MFLOPS calculation
    flops = flops + (rowsize + 2) * colsize * 2
END SUBROUTINE add_vec_vec

!------------------------------
! dot_product_vec_vec
!  IN  REAL(0:) v1  : vector 1
!  IN  REAL(0:) v2  : vector 2
!
!  RETURNS:
!    REAL: dot product
!
! TASK:
!   calculate dot product for vector 1 and vector 2
!
! GLOBAL DATA:
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT REAL    flops   : amount of FLOPS
!
! USED FUNCTIONS:
!   APPL:
!     --
!   MPI :
!     MPI_Allreduce
REAL FUNCTION dot_product_vec_vec(v1, v2)
    REAL, DIMENSION(0:), INTENT(IN) :: v1
    REAL, DIMENSION(0:), INTENT(IN) :: v2

    INTEGER :: i
    REAL    :: sum, global_sum

    ! sum = dot product of vectors
    sum = 0
    DO i = (rowsize + 2),((rowsize + 2) * (colsize + 1)) - 1
        sum = sum + v1(i) * v2(i)
    END DO

! global sum
#ifdef serial
    global_sum = sum
#endif
#ifndef serial
    commtime_start = MPI_WTIME()
    CALL MPI_ALLREDUCE(sum, global_sum, 1, MPI_REAL, MPI_SUM, comm_cart, ierror)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
    flops = flops + 1
#endif

    ! for the MFLOPS calculation
    flops = flops + (rowsize + 2) * colsize * 2

    dot_product_vec_vec = global_sum
END FUNCTION dot_product_vec_vec

!------------------------------
! max_norm_vec
!  IN  REAL(0:) v1  : vector 1
!
!  RETURNS:
!    REAL: maximum norm
!
! TASK:
!   calculate (vector maximum norm) for vector 1
!
! GLOBAL DATA:
!   IN INTEGER rowsize : number of data entries in a row of the local chunk
!   IN INTEGER colsize : number of data entries in a column of the local chunk
!
! USED FUNCTIONS:
!   APPL:
!     -- 
!   MPI :
!     MPI_ALLREDUCE
REAL FUNCTION max_norm_vec(v1)
    REAL, DIMENSION(0:), INTENT(IN) :: v1

    INTEGER :: i
    REAL    :: global_max = 0
    ! maxabs = maximum norm of the local vector
    REAL    :: maxabs = 0
    DO i = (rowsize + 2), ((rowsize + 2) * (colsize + 1)) - 1
       IF ( v1(i) > maxabs) THEN
            maxabs =  v1(i)
        END IF
        IF (-v1(i) > maxabs) THEN
            maxabs = -v1(i)
        END IF
    END DO

    ! global maximum
#ifdef serial
    global_max = maxabs
#endif
#ifndef serial
    commtime_start = MPI_WTIME()
    CALL MPI_ALLREDUCE(maxabs, global_max, 1, MPI_REAL, MPI_MAX, comm_cart, ierror)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
#endif
    ! no flops added, because we have only assignments and no operations
    max_norm_vec = global_max
END FUNCTION max_norm_vec

!------------------------------
! sqr_norm_vec
!  IN  REAL(0:) v1  : vector 1
!
!  RETURNS:
!    REAL: square of norm
!
! TASK:
!   calculate global (vector norm Euklid)^2 for vector 1
!
! GLOBAL DATA:
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT REAL    flops   : amount of FLOPS
!
! USED FUNCTIONS:
!   APPL:
!     -- 
!   MPI :
!     MPI_ALLREDUCE
REAL FUNCTION sqr_norm_vec(v1)
    REAL, DIMENSION(0:), INTENT(IN) :: v1

    INTEGER :: i
    REAL    :: sum, global_sum

    ! sum = dot product of vector
    sum = 0
    DO i = (rowsize + 2), ((rowsize + 2) * (colsize + 1)) - 1
        sum = sum + v1(i) * v1(i)
    END DO

    ! global sum
#ifdef serial
    global_sum = sum;
#endif
#ifndef serial
    commtime_start = MPI_WTIME()
    CALL MPI_Allreduce(sum, global_sum, 1, MPI_REAL, MPI_SUM, comm_cart, ierror)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
    flops = flops + 1
#endif
    ! for the MFLOPS calculation 
    flops = flops + (rowsize + 2) * colsize * 2
    sqr_norm_vec = global_sum
END FUNCTION sqr_norm_vec
!>>>>>>>>>>>>>>>>>>>>>> -end- of task 01 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 02 <<<<<<<<<<<<<<<<<<<<<<

!*********************** halo routines ************************

#ifndef serial
!------------------------------
!Init_Halo_Struct
!
! TASK:
!   Initialize the halo structure.
!
! GLOBAL DATA:
!   IN INTEGER rowsize : number of data entries in a row of the local chunk
!   IN INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT halo_structure halo : the halo
!   OUT REAL(0:) recv_scratch_buff_north : receive buffer for north halo
!   OUT REAL(0:) recv_scratch_buff_east  : receive buffer for east halo
!   OUT REAL(0:) recv_scratch_buff_south : receive buffer for south halo
!   OUT REAL(0:) recv_scratch_buff_west  : receive buffer for west halo
! USED FUNCTIONS:
!   APPL:
!     -- 
!   MPI:
!     MPI_CART_SHIFT
SUBROUTINE Init_Halo_Struct(v1)
    REAL, DIMENSION(0:), TARGET :: v1
    INTEGER                     :: maxScratchBuffSize

    ! allocate sufficient memory for the send scratch buffer
    IF (rowsize > colsize) THEN
        maxScratchBuffSize = rowsize
    ELSE
        maxScratchBuffSize = colsize
    END IF
    ! allocate sufficient memory for the receive scratch buffer for the halo 
    ALLOCATE(recv_scratch_buff_north(0:maxScratchBuffSize-1))
    ALLOCATE(recv_scratch_buff_east(0:maxScratchBuffSize-1))
    ALLOCATE(recv_scratch_buff_south(0:maxScratchBuffSize-1))
    ALLOCATE(recv_scratch_buff_west(0:maxScratchBuffSize-1))

    ! determine neighbors (1-dim and 2-dims)
    halo%north => v1(1:rowsize)
    halo%north_size = rowsize
    halo%south => v1((rowsize + 2) * (colsize + 1) + 1:(rowsize + 2) * (colsize + 1) + rowsize)
    halo%south_size = rowsize
    ! compute north_rank and south_rank
    CALL MPI_CART_SHIFT(comm_cart, 0, 1, halo%north_rank, halo%south_rank, ierror)

    halo%west => v1((rowsize + 2):(colsize * (rowsize + 2)):(rowsize + 2))
    halo%west_size = colsize
    halo%east => v1((rowsize + 2) * 2 - 1:((colsize + 1) * (rowsize + 2) - 1):(rowsize + 2))
    halo%east_size = colsize
    ! compute west_rank and east_rank
    CALL MPI_CART_SHIFT(comm_cart, 1, 1, halo%west_rank, halo%east_rank, ierror)
END SUBROUTINE Init_Halo_Struct

! halo communication: e.g. FROM_WEST_TO_EAST
!
! halo is included in the local vector chunk
!
!   **************************      **************************
!   *  -----NORTH HALO-----  *      *  -----NORTH HALO-----  *
!   *W ******************** E*      *W ******************** E*
!   *E *                *|* A*      *E *                  * A*
!   *S * WEST PROCESS   *------------> *   EAST PROCESS   * S*
!   *T *                *|* T*      *T *                  * T*
!   *  *                *|*  *      *  *                  *  *
!   *H ******************** H*      *H ******************** H*
!   *  -----SOUTH HALO-----  *      *  -----SOUTH HALO-----  *
!   **************************      **************************

!------------------------------
! comm_vec_halo
!
! IN  REAL(0:) v1: the vector with the data to send to other processes
!
! TASK:
!   Send neccessary data to the neighbors
!   and receive the needed data from the neighbors into
!   the halo area.
!
! USED GLOBAL DATA:
!   IN/OUT halo_structure halo : the halo
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   OUT REAL(0:) recv_scratch_buff_north : receive buffer for north halo
!   OUT REAL(0:) recv_scratch_buff_east  : receive buffer for east halo
!   OUT REAL(0:) recv_scratch_buff_south : receive buffer for south halo
!   OUT REAL(0:) recv_scratch_buff_west  : receive buffer for west halo
!
! USED FUNCTIONS:
!   MPI:
!     MPI_IRECV, MPI_SEND, MPI_WAITALL
!
! message tags for the different communication directions
SUBROUTINE comm_vec_halo(v1)
    REAL, DIMENSION(0:), INTENT(IN), TARGET :: v1
    INTEGER :: i

    INTEGER :: req_array(4)
    INTEGER :: status_array(MPI_STATUS_SIZE, 4)
    INTEGER :: TAG_FROM_NORTH_TO_SOUTH = 101
    INTEGER :: TAG_FROM_EAST_TO_WEST   = 102
    INTEGER :: TAG_FROM_SOUTH_TO_NORTH = 103
    INTEGER :: TAG_FROM_WEST_TO_EAST   = 104

    REAL, DIMENSION(:), POINTER :: send_to_east_values
    REAL, DIMENSION(:), POINTER :: send_to_west_values
    REAL, DIMENSION(:), POINTER :: send_to_north_values
    REAL, DIMENSION(:), POINTER :: send_to_south_values

    ! make a non-blocking receive

    ! Receive halo (non blocking). If current rank has no neighbors,
    ! the neighbor rank is MPI_PROC_NULL and nothing is sent.

    ! ATTENTION:
    ! It is NOT allowed to use pointers or strided arrays in NON-blocking send/receive routines,
    ! because the memory MAY NOT be contiguous. One must use a contiguous scratch buffer
    ! for sending and receiving, if the memory is not contiguous or a pointer to an array.

    ! receive north halo
    CALL MPI_IRECV(recv_scratch_buff_north, halo%north_size, MPI_REAL, halo%north_rank, &
                & TAG_FROM_NORTH_TO_SOUTH, comm_cart, req_array(1), ierror)
    ! receive east halo
    CALL MPI_IRECV(recv_scratch_buff_east, halo%east_size, MPI_REAL, halo%east_rank, &
                 & TAG_FROM_EAST_TO_WEST, comm_cart, req_array(2), ierror)
    ! receive south halo
    CALL MPI_IRECV(recv_scratch_buff_south, halo%south_size, MPI_REAL, halo%south_rank, &
                 & TAG_FROM_SOUTH_TO_NORTH, comm_cart, req_array(3), ierror)
    ! receive west halo
    CALL MPI_IRECV(recv_scratch_buff_west, halo%west_size, MPI_REAL, halo%west_rank, &
                & TAG_FROM_WEST_TO_EAST, comm_cart, req_array(4), ierror)
   
    ! send to north neighbor (blocking)
    send_to_north_values => v1(SIDX(1, 1):SIDX(colsize, 1):1)
    CALL MPI_SEND(send_to_north_values, halo%north_size, MPI_REAL, halo%north_rank, &
                & TAG_FROM_SOUTH_TO_NORTH, comm_cart, ierror)

    ! send to east neighbor (blocking)
    send_to_east_values => v1(SIDX(1, rowsize):SIDX(colsize, rowsize):rowsize+2)
    CALL MPI_SEND(send_to_east_values, halo%east_size, MPI_REAL, halo%east_rank, &
                & TAG_FROM_WEST_TO_EAST, comm_cart, ierror)

    ! send to south neighbor (blocking)
    send_to_south_values => v1(SIDX(colsize, 1):SIDX(colsize, rowsize):1)
    CALL MPI_SEND(send_to_south_values, halo%south_size, MPI_REAL, halo%south_rank, &
                & TAG_FROM_NORTH_TO_SOUTH, comm_cart, ierror)
    
    ! send to west neighbor (blocking)
    send_to_west_values => v1(SIDX(1, 1):SIDX(colsize, 1):rowsize+2)
    CALL MPI_SEND(send_to_west_values, halo%west_size, MPI_REAL, halo%west_rank, &
                & TAG_FROM_EAST_TO_WEST, comm_cart, ierror)

    ! wait for all halo info
    CALL MPI_WAITALL(4, req_array, status_array, ierror)

    ! copy the received halo from the scratch buff into the real halo
    ! ATTENTION: pointers always start with the index 1!!!
    IF (halo%north_rank /= MPI_PROC_NULL) THEN
        DO i = 0, halo%north_size - 1
            halo%north(i+1) = recv_scratch_buff_north(i)
        END DO
    END IF
    IF (halo%east_rank /= MPI_PROC_NULL) THEN
        DO i = 0, halo%east_size - 1
            halo%east(i+1) = recv_scratch_buff_east(i)
        END DO
    END IF
    IF (halo%south_rank /= MPI_PROC_NULL) THEN
        DO i = 0, halo%south_size - 1
            halo%south(i+1) = recv_scratch_buff_south(i)
        END DO
    END IF
    IF (halo%west_rank /= MPI_PROC_NULL) THEN
        DO i = 0, halo%west_size - 1
            halo%west(i+1) = recv_scratch_buff_west(i)
        END DO
    END IF
END SUBROUTINE comm_vec_halo
#endif

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 02 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 03 <<<<<<<<<<<<<<<<<<<<<<

!********************** matrix-vector-multiply ****************************

!------------------------------
! mult_A_vec
!      IN  REAL(0:) v1 : vector to multiply A with
!      OUT REAL(0:) v2 : output vector
!
! TASK: Calculates matrix vector product v2 = A * v1
!
! USED GLOBAL DATA: 
!   IN     INTEGER rowsize: number of data entries in a row of the local chunk
!   IN     INTEGER colsize: number of data entries in a column of the local chunk
!   IN/OUT REAL    flops  : amount of FLOPS
!
! USED FUNCTIONS:
!   --
SUBROUTINE mult_A_vec(v1, v2)
    REAL, DIMENSION(0:), INTENT(IN)  :: v1
    REAL, DIMENSION(0:), INTENT(OUT) :: v2
    INTEGER :: i, j

#ifdef FASTMEMXS
    INTEGER :: k, calcend
#endif

    !**************************************************************************
    !* Matrix-Vector mult. without distinction of border/halo neighbored data *
    !**************************************************************************

#ifdef FASTMEMXS
    ! multiplication only for the inner values
    calcend = (colsize + 1) - 2 - 1
    IF (MODULO(calcend, 2) == 0) THEN
        calcend = calcend - 1
    END IF
    DO i = 1, calcend, 2
        DO j = 1, (rowsize + 1) - 1
            v2(SIDX(i, j)) = -1 * v1(SIDX((i - 1), j)) &
                          & - 1 * v1(SIDX(i, (j - 1))) &  
                          & + 4 * v1(SIDX(i, j)) &
                          & - 1 * v1(SIDX(i, (j + 1))) &
                          & - 1 * v1(SIDX((i + 1), j))
            v2(SIDX(i + 1, j)) = -1 * v1(SIDX((i - 1 + 1), j)) &
                          & - 1 * v1(SIDX((i + 1), (j - 1))) &  
                          & + 4 * v1(SIDX((i + 1), j)) &
                          & - 1 * v1(SIDX((i + 1), (j + 1))) &
                          & - 1 * v1(SIDX((i + 1 + 1), j))
        END DO
    END DO
 
    ! calculate the remaining rows
    DO k = calcend + 2, (colsize + 1) - 1
        DO j = 1, (rowsize + 1) - 1
            v2(SIDX(k, j)) = -1 * v1(SIDX((k - 1), j)) &
                          & - 1 * v1(SIDX(k, (j - 1))) &
                          & + 4 * v1(SIDX(k, j)) &
                          & - 1 * v1(SIDX(k, (j + 1))) &
                          & - 1 * v1(SIDX((k + 1), j))
        END DO
    END DO
#endif
#ifndef FASTMEMXS
    ! multiplication only for the inner values
    DO i = 1, (colsize + 1) - 1
        DO j = 1, (rowsize + 1) - 1
            v2(SIDX(i, j)) = -1 * v1(SIDX((i - 1), j)) &
                          & - 1 * v1(SIDX(i, (j - 1))) &  
                          & + 4 * v1(SIDX(i, j)) &
                          & - 1 * v1(SIDX(i, (j + 1))) &
                          & - 1 * v1(SIDX((i + 1), j))
        END DO
    END DO
#endif

    ! for the MFLOPS calculation
    flops = flops + rowsize * colsize * 9
END SUBROUTINE mult_A_vec

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 03 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 04 <<<<<<<<<<<<<<<<<<<<<<

!******************* domain decomposition **********************

!----------------------- 1-dim decomposition -----------------
!
!   Example: m=6, n=7, numprocs=2
!
!       |       j=0     1     2     3     4     5     6 |           
!   ----+-----------------------------------------------+-----------
!   i=0 |         0     1     2     3     4     5     6 |           
!     1 |         7     8     9    10    11    12    13 | process 1 
!     2 |        14    15    16    17    18    19    20 |           
!   ----+-----------------------------------------------+-----------
!     3 |        21    22    23    24    25    26    27 |           
!     4 |        28    29    30    31    32    33    34 | process 2 
!     5 |        35    36    37    38    39    40    41 |           
!   ----+-----------------------------------------------+-----------

!------------------------------
! Init_1dim
!
! TASK:
!   Do the one dimensional domain decomposition as depicted above.
!
! USED GLOBAL DATA:
!   OUT INTEGER start_i    : start row of the physical area this process is to calculate
!   OUT INTEGER end1_i     : end row + 1 of the physical area this process is to calculate
!   OUT INTEGER start_j    : start column of the physical area this process is to calculate
!   OUT INTEGER end1_j     : end column + 1 of the physical area this process is to calculate
!   IN/OUT INTEGER my_rank : the rank of the current process
!   OUT INTEGER comm_cart  : cartesian communicator
!   IN  INTEGER numprocs   : number of used processors
!   IN  INTEGER m          : vertical size of the physical problem
!   IN  INTEGER n          : horizontal size of the physical problem
!
! USED FUNCTIONS:
!   MPI:
!     MPI_CART_CREATE, MPI_COMM_RANK
SUBROUTINE Init_1dim()
    INTEGER :: slice_size
#ifndef serial
    INTEGER :: dims(0:1)
    LOGICAL :: periods(0:1)
    IF ((print_level >= 1) .AND. (my_rank == 0)) THEN
        WRITE(*,'(A, I3)') 'Using 1-dim domain decomposition, numprocs=', numprocs
    END IF
    ! Init the virtual topology (1-dim)
    dims(0) = numprocs
    dims(1) = 1
    periods(0) = .FALSE.
    periods(1) = .FALSE.


#endif

    ! the size of each slice
    ! If m is not a multiple of numprocs, we round up the slice size
    slice_size = ((m - 1) / numprocs) + 1

    ! the rows of the physical area are distributed in slices of equal size
    ! (maybe except for the last slice) to each process
    start_i = (my_rank * slice_size)
    IF (start_i > m) THEN
        start_i = m
    END IF

    end1_i = start_i + slice_size
    IF (end1_i > m) THEN
        end1_i = m
    END IF

    ! we have the full vertical extent (all n columns) for each row 
    start_j = 0
    end1_j = n
END SUBROUTINE Init_1dim

!----------------------- 2-dim decomposition -----------------
!
!   Example: m=5, n=7, m_procs=2, n_procs=2 
! 
!       |           j=0     1     2     3  |  4     5     6            |
!   ----+----------------------------------+---------------------------+
!   i=0 |             0     1     2     3  |  4     5     6            |
!     1 | process 1   7     8     9    10  | 11    12    13  process 2 |
!     2 |            14    15    16    17  | 18    19    20            |
!   ----+----------------------------------+---------------------------+
!     3 | process 3  21    22    23    24  | 25    26    27  process 4 |
!     4 |            28    29    30    31  | 32    33    34            |
!   ----+------------------------------+-------------------------------+

!------------------------------
! Init_2dim
!
! TASK:
!   Do the two dimensional domain decomposition as depicted above.
!
! USED GLOBAL DATA:
!   OUT INTEGER start_i   : start row of the physical area this process is to calculate
!   OUT INTEGER end1_i    : end row + 1 of the physical area this process is to calculate
!   OUT INTEGER start_j   : start column of the physical area this process is to calculate
!   OUT INTEGER end1_j    : end column + 1 of the physical area this process is to calculate
!   IN  INTEGER my_rank   : the rank of the current process
!   IN  INTEGER m_procs   : vertical number of processors
!   IN  INTEGER n_procs   : horizontal number of processors
!   OUT INTEGER comm_cart : cartesian communicator
!   IN  INTEGER m         : vertical size of the physical problem
!   IN  INTEGER n         : horizontal size of the physical problem
!
! USED FUNCTIONS:
!   APPL:
!     abrt
!   MPI:
!     MPI_CART_CREATE, MPI_COMM_RANK
SUBROUTINE Init_2dim()
    INTEGER :: blocksize_i, blocksize_j
    INTEGER :: startblock_i, startblock_j
#ifndef serial
    INTEGER :: dims(0:1)
    LOGICAL :: periods(0:1)
    IF ((print_level >= 1) .AND. (my_rank == 0)) THEN
        WRITE(*,'(A, I3, A, I3, A, I3)') 'Using 2-dim domain decomposition, m_procs=', m_procs, ', &
                 & n_procs=', n_procs, ', numprocs=', numprocs
    END IF
    ! Init the virtual topology (2-dims)
    dims(0) = m_procs
    dims(1) = n_procs
    periods(0) = .FALSE.
    periods(1) = .FALSE.


#endif

    IF ((my_rank == 0) .AND. ((m_procs * n_procs) /= numprocs)) THEN
        WRITE(*,*) 'Number of processors does not fit! Use ', (m_procs * n_procs), ' processors instead. &
             & Aborting...'
        CALL abrt()
    END IF

    ! the number of rows for each block (the last block might have
    ! less than the first blocks)


    ! the number of columns for each block (the last block might have
    ! less than the first blocks)


    ! calculate the start and end row and column












END SUBROUTINE Init_2dim

!------------------------------
! Init_Decomp
!
! TASK:
!   Do the domain decomposition.
!
! USED GLOBAL DATA:
!   IN/OUT INTEGER rowsize     : number of data entries in a row of the local chunk
!   IN/OUT INTEGER colsize     : number of data entries in a column of the local chunk
!   IN/OUT INTEGER chunksize   : number of data entries in the local chunk PLUS halo data
!   IN     INTEGER decomp_dims : dimension of the domain decomposition 
!   IN     INTEGER my_rank     : the rank of the current process
!
! USED FUNCTIONS:
!   APPL:
!     abrt, Init_1dim, Init_2dim
SUBROUTINE Init_Decomp()
    SELECT CASE (decomp_dims)    
        CASE (1)
            CALL Init_1dim()
        CASE (2)
            CALL Init_2dim()
        CASE DEFAULT
            WRITE(*,*) 'Wrong decomp_dims=', decomp_dims
            CALL abrt()
    END SELECT
    ! calculate the size of the rows, cols and the whole chunk
    rowsize   = end1_j - start_j
    colsize   = end1_i - start_i
    chunksize = (rowsize + 2) * (colsize + 2)

    IF ((rowsize < 2) .OR. (colsize < 2)) THEN
        WRITE(*,*) 'Each process must have at least 2 entries in each dimension.'
        WRITE(*,*) 'Decrease amount of processes or increase physical area size.'
        WRITE(*, '(A, I3, A, I0, A, I0)') 'Rank[', my_rank, ']: rowsize = ', rowsize, ' -- colsize = ', colsize
        WRITE(*,*) 'Aborting.'
        CALL abrt()
    END IF
END SUBROUTINE Init_Decomp

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 04 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 05 <<<<<<<<<<<<<<<<<<<<<<
!************* boundary and exact solution init ****************

!------------------------------
! Init_Boundary_Vec
!  OUT REAL(0:) b  : the boundary vector 
!
! TASK: Store all boundary values in the vector memory
!       (that is already allocated).
!       In this example, the boundaries are given by:
!         u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=-1,m and j=-1,n
!
! GLOBAL DATA:
!   IN INTEGER start_i : start row of the physical area this process is to calculate
!   IN INTEGER end1_i  : end row + 1 of the physical area this process is to calculate
!   IN INTEGER start_j : start column of the physical area this process is to calculate
!   IN INTEGER end1_j  : end column + 1 of the physical area this process is to calculate
!   IN INTEGER m       : vertical size of the physical problem
!   IN INTEGER n       : horizontal size of the physical problem
!
! USED FUNCTIONS:
!     --
SUBROUTINE Init_Boundary_Vec(b)
    REAL, DIMENSION(0:), INTENT(OUT) :: b
    INTEGER                          :: i, j
    REAL                             :: v, h

    h = 1.0/(m+1)
    ! boundary vector data is all zero at this point
    DO i = start_i, end1_i - 1
        DO j = start_j, end1_j - 1
            v = 0.0
            ! next line not neccessary, because boundary value is zero
            ! if (i==0)   v = v + u(-1,j) = 0
            ! next line simplified to add 1.0 to vector, because of the boundary value of 1.0
            ! if (i==m-1) v = v + u(m, j): h * (m+1) = 1
            IF (i==m-1) THEN
                v = 1.0
            END IF
            ! u(i,-1):  h * (i+1)
            IF (j==0) THEN
                v = v + h * (i + 1)
            END IF
            ! u(i, n):  h * (i+1)
            IF (j==n-1) THEN
                v = v + h * (i + 1)
            END IF
            b(SIDX((i - start_i + 1), (j - start_j + 1))) = v
        END DO
    END DO
END SUBROUTINE Init_Boundary_Vec

!------------------------------
! Init_Exact_Solution_Vec
!   OUT REAL(0:) u  : the local chunk of the exact solution vector
!
! TASK: Store the exact solution for the given boundary in the vector u.
!       In this example, the inner values are given by:
!       u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=0..m-1 and j=0..n-1
!
! GLOBAL DATA:
!   IN INTEGER start_i : start row of the physical area this process is to calculate
!   IN INTEGER end1_i  : end row + 1 of the physical area this process is to calculate
!   IN INTEGER start_j : start column of the physical area this process is to calculate
!   IN INTEGER end1_j  : end column + 1 of the physical area this process is to calculate
!   IN INTEGER m       : vertical size of the physical problem
!
! USED FUNCTIONS:
!   --
SUBROUTINE Init_Exact_Solution_vec(u)
    REAL, DIMENSION(0:), INTENT(OUT) :: u
    INTEGER                          :: i, j
    REAL                             :: h

    h = 1.0 / (m + 1)
    DO i = start_i, end1_i - 1
        DO j = start_j, end1_j - 1
            u(SIDX((i - start_i + 1), (j - start_j + 1))) = h * (i + 1)
        END DO
    END DO
END SUBROUTINE Init_Exact_Solution_vec
!>>>>>>>>>>>>>>>>>>>>>> -end- of task 05 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 06 <<<<<<<<<<<<<<<<<<<<<<

!******************* print routines ********************************
SUBROUTINE appl_print_vec(name, v, mprint, nprint)
    CHARACTER, INTENT(IN)              :: name
    REAL, DIMENSION(0:,0:), INTENT(IN) :: v
    INTEGER, INTENT(IN)                :: mprint, nprint
    INTEGER                            :: i,j
    IF (my_rank == 0) THEN
        WRITE(*,'(A, A)', ADVANCE='NO') name, ':       j='
        DO j=0, n - 1 
            IF ( (n <= nprint) .OR. ( (n > nprint) .AND. ((nprint-1)*(j+n-2)/(n-1) /= (nprint-1)*(j+n-1)/(n-1)))) THEN
                WRITE(*,'(I6)', ADVANCE='NO') j
            END IF
        END DO
        WRITE(*,*)
        DO i=0, m - 1
            IF ( (m <= mprint) .OR. ( (m > mprint) .AND. ((mprint-1)*(i+m-2)/(m-1) /= (mprint-1)*(i+m-1)/(m-1))  ) ) THEN
                WRITE(*,'(A, A, I3, A)', ADVANCE='NO') name, ': i=', i, '   '
                DO j=0, n - 1
                   IF ( (n <= nprint) .OR. ((n > nprint) .AND. ((nprint-1)*(j+n-2)/(n-1) /= (nprint-1)*(j+n-1)/(n-1)))) THEN
                       WRITE(*, '(A, F5.3)', ADVANCE='NO') ' ', v(i, j)
                   END IF
                END DO
                WRITE(*,*)
            END IF
        END DO
    END IF
END SUBROUTINE appl_print_vec

SUBROUTINE recv_token()
#ifndef serial
    INTEGER :: dummy, status(MPI_STATUS_SIZE)
    IF (my_rank > 0) THEN
        CALL MPI_RECV(dummy, 0, MPI_INTEGER, my_rank-1, 333, comm_cart, status, ierror)
    END IF
#endif
END SUBROUTINE recv_token

SUBROUTINE flush_and_send_token()
#ifndef serial
    INTEGER    :: dummy, status(MPI_STATUS_SIZE)
    CALL MPI_SEND(dummy, 0, MPI_INTEGER, MODULO((my_rank+1), numprocs), 333, comm_cart, ierror)
    ! Process 0 must wait until the last process (numprocs-1) has finished its output
    IF (my_rank == 0) THEN
        CALL MPI_RECV(dummy, 0, MPI_INTEGER, numprocs-1, 333, comm_cart, status, ierror)
    END IF
#endif
END SUBROUTINE flush_and_send_token

SUBROUTINE print_halo()
#ifndef serial
    INTEGER :: i

    CALL recv_token()

    WRITE(*,*)
    WRITE(*,*) 'Halo of rank ', my_rank
    WRITE(*,*) '----------------'
    IF (halo%north_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'North: NO halo'
    ELSE
        WRITE(*,*) 'North:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%north_size, ', neighbor rank: ', halo%north_rank
        DO i = 1, halo%north_size
            WRITE(*,*) 'halo[', i, ']: ', halo%north(i)
        END DO
    END IF
    IF (halo%east_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'East: NO halo'
    ELSE
        WRITE(*,*) 'East:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%east_size, ', neighbor rank: ', halo%east_rank
        DO i = 1, halo%east_size
            WRITE(*,*) 'halo[', i, ']: ', halo%east(i)
        END DO
    END IF
    IF (halo%south_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'South: NO halo'
    ELSE
        WRITE(*,*) 'South:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%south_size, ', neighbor rank: ', halo%south_rank
        DO i = 1, halo%south_size
            WRITE(*,*) 'halo[', i, ']: ', halo%south(i)
        END DO
    END IF
    IF (halo%west_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'West: NO halo'
    ELSE
        WRITE(*,*) 'West:'
        WRITE(*,*) '------'
        WRITE(*,*) 'values: ', halo%west_size, ', neighbor rank: ', halo%west_rank
        DO i = 1, halo%west_size
            WRITE(*,*) 'halo[', i, ']: ', halo%west(i)
        END DO
    END IF
    WRITE(*,*) ''
    CALL flush_and_send_token()
#endif
END SUBROUTINE print_halo

SUBROUTINE print_halo_struct()
#ifndef serial
    CALL recv_token()

    WRITE(*,*) ''
    WRITE(*,*) 'Halo of rank ', my_rank
    WRITE(*,*) '----------------'
    IF (halo%north_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'North: NO halo'
    ELSE
        WRITE(*,*) 'North: values: ', halo%north_size ,', neighbor rank: ', halo%north_rank
    END IF
    IF (halo%east_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'East: NO halo'
    ELSE
        WRITE(*,*) 'East: values: ', halo%east_size ,', neighbor rank: ', halo%east_rank
    END IF
    IF (halo%south_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'South: NO halo'
    ELSE
        WRITE(*,*) 'South: values: ', halo%south_size ,', neighbor rank: ', halo%south_rank
    END IF
    IF (halo%west_rank == MPI_PROC_NULL) THEN
        WRITE(*,*) 'West: NO halo'
    ELSE
        WRITE(*,*) 'West: values: ', halo%west_size ,', neighbor rank: ', halo%west_rank
    END IF
    WRITE(*,*) ''

    CALL flush_and_send_token()
#endif
END SUBROUTINE print_halo_struct

SUBROUTINE print_vec2d(name, v2, mprint, nprint)
    CHARACTER, INTENT(IN)           :: name
    REAL, DIMENSION(0:), INTENT(IN) :: v2
    INTEGER, INTENT(IN)             :: mprint, nprint
    INTEGER :: i,j
    REAL    :: whole_v2(0:m-1, 0:n-1)
#ifndef serial
    INTEGER    :: rank
    REAL       :: buff(0:(n + 2) * (m + 2) - 1)
    INTEGER    :: extent(0:3), new_offset
    INTEGER    :: status(MPI_STATUS_SIZE), status2(MPI_STATUS_SIZE)

    IF (my_rank > 0) THEN
        ! send the start_i, end1-i, start_j, end1_j
        extent(0) = start_i
        extent(1) = end1_i
        extent(2) = start_j
        extent(3) = end1_j
        CALL MPI_SEND(extent, 4, MPI_INTEGER, 0, 666, comm_cart, ierror)

        ! send the vector data to rank 0
        CALL MPI_SEND(v2, chunksize, MPI_REAL, 0, 333, comm_cart, ierror)
    ELSE
        ! copy the rank 0 chunk into the whole physical vector
        DO i = start_i, end1_i - 1
            DO j = start_j, end1_j - 1
                whole_v2(i, j) = v2(SIDX((i - start_i + 1), (j - start_j + 1)))
            END DO
        END DO
            
        ! receive the halo data and data extent from all other ranks into the buffer
        ! and copy the buffer to where it belongs into the whole physical vector
        DO rank = 1, numprocs - 1
            CALL MPI_RECV(extent, 4, MPI_INTEGER, rank, 666, comm_cart, status, ierror)
            CALL MPI_RECV(buff, ((extent(1) - extent(0) + 2) * (extent(3) - extent(2) + 2)), MPI_REAL, &
                        & rank, 333, comm_cart, status2, ierror)
    
            new_offset = extent(3) - extent(2) + 2
            
            DO i = extent(0), extent(1) - 1
                DO j = extent(2), extent(3) - 1
                    whole_v2(i, j) = buff((i - extent(0) + 1) * new_offset + (j - extent(2) + 1))
                END DO
            END DO
        END DO
    END IF
    
    CALL appl_print_vec(name, whole_v2, mprint, nprint)
#endif
#ifdef serial
    ! copy the data without halo into the whole physical vector
    DO i = start_i, end1_i - 1
        DO j = start_j, end1_j - 1
            whole_v2(i, j) = v2(SIDX((i - start_i + 1), (j - start_j + 1)))
        END DO
    END DO
    CALL appl_print_vec(name, whole_v2, mprint, nprint)
#endif
END SUBROUTINE print_vec2d

SUBROUTINE print_vec(name, v1)
    CHARACTER, INTENT(IN)           :: name
    REAL, DIMENSION(0:), INTENT(IN) :: v1
    INTEGER                         :: i, j
    CALL recv_token()
    DO i = 1, (colsize + 1) - 1
        DO j = 1, (rowsize + 1) - 1
            WRITE(*,'(A, A, I2, A, F6.3)') name, '[', ((i * rowsize) + j), '] = ', v1(SIDX(i, j))
        END DO
    END DO
    CALL flush_and_send_token()
END SUBROUTINE print_vec

SUBROUTINE print_decomp()
    CALL recv_token()
    WRITE(*, '(A,I3, A, I0, A, I0, A, I0, A, I0)') 'Rank ', my_rank, ': start_i, end1_i, start_j, end1_j: ', &
             & start_i, ', ', end1_i, ', ', start_j, ', ', end1_j
    CALL flush_and_send_token()
END SUBROUTINE print_decomp

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 06 <<<<<<<<<<<<<<<<<<<<<<
!>>>>>>>>>>>>>>>>>>>>>> begin of task 07 <<<<<<<<<<<<<<<<<<<<<<

!************************ CG solver ****************************

!------------------------------
! CG_solve
!   OUT INTEGER num_iter : return the number of iterations
!   OUT REAL    norm_r   : return the norm of r
!
! TASK:
!   solve Laplace equation
!
! GLOBAL DATA:
!   IN     INTEGER     iter_max : maximal number of iterations
!   IN     REAL        epsilon  : epsilon for maximal error
!   IN     REAL(0:)    b        : local chunk of vector b
!   IN/OUT REAL(0:)    x        : local chunk of vector x
!   IN     INTEGER      my_rank  : rank of the local process
! USED FUNCTIONS:
!   APPL:
!     comm_vec_halo, print_vec, add_vec_vec
!     sqr_norm_vec, print_halo, mult_A_vec, dot_product_vec_vec
!   MPI-lib:
!     MPI_WTIME
SUBROUTINE CG_solve(num_iter, norm_r)
    INTEGER, INTENT(INOUT) :: num_iter
    REAL, INTENT(INOUT)    :: norm_r

    ! only p has a halo
    REAL, DIMENSION(0:chunksize-1) :: p, r, v
    REAL                           :: alpha, alpha_new, lambda
    INTEGER                        :: iter

    ! set the content of the vectors to zero
    p = 0
    r = 0
    v = 0

#ifndef serial
        ! determine communication partners for vector p !!! 
    CALL Init_Halo_Struct(p)
    IF (print_level>=4) THEN
        CALL print_halo_struct()
    END IF
#endif

    ! x is initialized with 0

    ! instead of p = Ax
    ! p := x
    p = x

#ifndef serial
    ! update halo
    commtime_start = MPI_WTIME()
    CALL comm_vec_halo(p)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
#endif

    ! v = Ap
    CALL mult_A_vec(p, v)
    IF (print_level>=4) THEN
        CALL print_vec('v', v)
    END IF

    ! p = b - v
    CALL add_vec_vec(b, -1.0, v, p)
    IF (print_level>=4) THEN
        CALL print_vec('p', p)
    END IF

    ! r := p
    r = p
    IF (print_level>=4) THEN
        CALL print_vec('r', r)
    END IF

    ! alpha = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2
    alpha = sqr_norm_vec(r)

    iter = 1
    DO WHILE ((iter < iter_max) .AND. (alpha > epsilon))
        iter = iter + 1
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            WRITE(*,'(A, I3, A, I0, A, F9.6)') 'CG - my_rank =', my_rank, ', i=', iter, ', alpha=', alpha
        END IF

#ifndef serial
        ! update halo
        commtime_start = MPI_WTIME()
        CALL comm_vec_halo(p)
        commtime_end = MPI_WTIME()
        commtime_sum = commtime_sum + commtime_end - commtime_start
#endif

        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('p', p)
            CALL print_halo()
        END IF

        ! v = Ap
        CALL mult_A_vec(p, v)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('v', v) 
        END IF

        ! lambda = alpha / (v, p)2 
        lambda = alpha / dot_product_vec_vec(v, p)

        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            WRITE(*,'(A, I3, A, I0, A, F9.6, A, F9.6)') 'CG - my_rank =', my_rank, ', i=', iter, ', vp=', &
                     & dot_product_vec_vec(v, p), ', lambda=', lambda
        END IF

        ! x := x + lambda * p
        CALL add_vec_vec(x, lambda, p, x)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('x', x)
        END IF

        ! r := r - lambda * v
        CALL add_vec_vec(r, -lambda, v, r)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('r', r)
        END IF

        ! alpha_new = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2 
        alpha_new = sqr_norm_vec(r)

        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            WRITE(*,'(A, I3, A, I0, A, F9.6)') 'CG - my_rank =', my_rank, ', i=', iter, ', alpha_new=', alpha_new
        END IF

        ! p := r + (alpha_new/alpha) * p 
        CALL add_vec_vec(r, alpha_new/alpha, p, p)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('p', p)
        END IF

        ! alpha := alpha_new
        alpha = alpha_new

        IF ((print_level>=2) .AND. (my_rank==0)) THEN
            WRITE (*,'(A, I0, A, E8.2, A)', ADVANCE='NO') 'CG i=', iter, ', ||r||**2 = ', alpha_new, ' '
            IF (alpha_new <= epsilon) THEN
                WRITE (*,'(A)', ADVANCE='NO') '<='
            ELSE
                WRITE (*,'(A)', ADVANCE='NO') '>' 
            ENDIF
            WRITE (*,'(A, E8.2, A, E8.2, A)', ADVANCE='NO') ' ', epsilon, ',   ||r|| = ', sqrt(alpha_new), ' '
            IF (SQRT(alpha_new) <= SQRT(epsilon)) THEN
                WRITE (*,'(A)', ADVANCE='NO') '<='
            ELSE
                WRITE (*,'(A)', ADVANCE='NO') '>' 
            ENDIF
            WRITE (*,'(A, E8.2)') ' ', sqrt(epsilon)
        END IF
    END DO

    ! update output-arguments
    num_iter = iter
    norm_r = SQRT(alpha)
END SUBROUTINE CG_solve

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 07 <<<<<<<<<<<<<<<<<<<<<<
END PROGRAM cgv_fortran

