! CGV/solution/parts/cgvp01.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
! >>>>>>>>>>>>>>>>>>>>>> begin of task 01 <<<<<<<<<<<<<<<<<<<<<<

! ********************** vector routines ************************

!-------------------------------------------
! add_vec_vec
!  IN  REAL(0:) v1      : vector 1
!  IN  REAL     lambda  : factor lambda
!  IN  REAL(0:) v2      : vector 2
!  OUT REAL(0:) v3      : vector 3
!
! RETURNS:
!   --
!
! TASK:
!   calculate:   v3 := v1 + lambda * v2
!
! GLOBAL DATA:
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT REAL    flops   : amount of FLOPS
!
! USED FUNCTIONS:
!   --
SUBROUTINE add_vec_vec(v1, lambda, v2, v3)
    REAL, DIMENSION(0:), INTENT(IN)  ::  v1
    REAL, INTENT(IN)                 ::  lambda
    REAL, DIMENSION(0:), INTENT(IN)  ::  v2
    REAL, DIMENSION(0:), INTENT(OUT) ::  v3
    INTEGER                          ::  i

    ! addition of two vectors, second vector is multiplied by a scalar
    DO i = (rowsize + 2), ((rowsize + 2) * (colsize + 1)) - 1
        v3(i) = v1(i) + lambda * v2(i);
    END DO

    ! for the MFLOPS calculation
    flops = flops + (rowsize + 2) * colsize * 2
END SUBROUTINE add_vec_vec

!------------------------------
! dot_product_vec_vec
!  IN  REAL(0:) v1  : vector 1
!  IN  REAL(0:) v2  : vector 2
!
!  RETURNS:
!    REAL: dot product
!
! TASK:
!   calculate dot product for vector 1 and vector 2
!
! GLOBAL DATA:
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT REAL    flops   : amount of FLOPS
!
! USED FUNCTIONS:
!   APPL:
!     --
!   MPI :
!     MPI_Allreduce
REAL FUNCTION dot_product_vec_vec(v1, v2)
    REAL, DIMENSION(0:), INTENT(IN) :: v1
    REAL, DIMENSION(0:), INTENT(IN) :: v2

    INTEGER :: i
    REAL    :: sum, global_sum

    ! sum = dot product of vectors
    sum = 0
    DO i = (rowsize + 2),((rowsize + 2) * (colsize + 1)) - 1
        sum = sum + v1(i) * v2(i)
    END DO

! global sum
#ifdef serial
    global_sum = sum
#endif
#ifndef serial
    commtime_start = MPI_WTIME()
    CALL MPI_ALLREDUCE(sum, global_sum, 1, MPI_REAL, MPI_SUM, comm_cart, ierror)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
    flops = flops + 1
#endif

    ! for the MFLOPS calculation
    flops = flops + (rowsize + 2) * colsize * 2

    dot_product_vec_vec = global_sum
END FUNCTION dot_product_vec_vec

!------------------------------
! max_norm_vec
!  IN  REAL(0:) v1  : vector 1
!
!  RETURNS:
!    REAL: maximum norm
!
! TASK:
!   calculate (vector maximum norm) for vector 1
!
! GLOBAL DATA:
!   IN INTEGER rowsize : number of data entries in a row of the local chunk
!   IN INTEGER colsize : number of data entries in a column of the local chunk
!
! USED FUNCTIONS:
!   APPL:
!     -- 
!   MPI :
!     MPI_ALLREDUCE
REAL FUNCTION max_norm_vec(v1)
    REAL, DIMENSION(0:), INTENT(IN) :: v1

    INTEGER :: i
    REAL    :: global_max = 0
    ! maxabs = maximum norm of the local vector
    REAL    :: maxabs = 0
    DO i = (rowsize + 2), ((rowsize + 2) * (colsize + 1)) - 1
       IF ( v1(i) > maxabs) THEN
            maxabs =  v1(i)
        END IF
        IF (-v1(i) > maxabs) THEN
            maxabs = -v1(i)
        END IF
    END DO

    ! global maximum
#ifdef serial
    global_max = maxabs
#endif
#ifndef serial
    commtime_start = MPI_WTIME()
    CALL MPI_ALLREDUCE(maxabs, global_max, 1, MPI_REAL, MPI_MAX, comm_cart, ierror)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
#endif
    ! no flops added, because we have only assignments and no operations
    max_norm_vec = global_max
END FUNCTION max_norm_vec

!------------------------------
! sqr_norm_vec
!  IN  REAL(0:) v1  : vector 1
!
!  RETURNS:
!    REAL: square of norm
!
! TASK:
!   calculate global (vector norm Euklid)^2 for vector 1
!
! GLOBAL DATA:
!   IN     INTEGER rowsize : number of data entries in a row of the local chunk
!   IN     INTEGER colsize : number of data entries in a column of the local chunk
!   IN/OUT REAL    flops   : amount of FLOPS
!
! USED FUNCTIONS:
!   APPL:
!     -- 
!   MPI :
!     MPI_ALLREDUCE
REAL FUNCTION sqr_norm_vec(v1)
    REAL, DIMENSION(0:), INTENT(IN) :: v1

    INTEGER :: i
    REAL    :: sum, global_sum

    ! sum = dot product of vector
    sum = 0
    DO i = (rowsize + 2), ((rowsize + 2) * (colsize + 1)) - 1
        sum = sum + v1(i) * v1(i)
    END DO

    ! global sum
#ifdef serial
    global_sum = sum;
#endif
#ifndef serial
    commtime_start = MPI_WTIME()
    CALL MPI_Allreduce(sum, global_sum, 1, MPI_REAL, MPI_SUM, comm_cart, ierror)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
    flops = flops + 1
#endif
    ! for the MFLOPS calculation 
    flops = flops + (rowsize + 2) * colsize * 2
    sqr_norm_vec = global_sum
END FUNCTION sqr_norm_vec
!>>>>>>>>>>>>>>>>>>>>>> -end- of task 01 <<<<<<<<<<<<<<<<<<<<<<
