! CGV/solution/parts/cgvp04.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
!>>>>>>>>>>>>>>>>>>>>>> begin of task 04 <<<<<<<<<<<<<<<<<<<<<<

!******************* domain decomposition **********************

!----------------------- 1-dim decomposition -----------------
!
!   Example: m=6, n=7, numprocs=2
!
!       |       j=0     1     2     3     4     5     6 |           
!   ----+-----------------------------------------------+-----------
!   i=0 |         0     1     2     3     4     5     6 |           
!     1 |         7     8     9    10    11    12    13 | process 1 
!     2 |        14    15    16    17    18    19    20 |           
!   ----+-----------------------------------------------+-----------
!     3 |        21    22    23    24    25    26    27 |           
!     4 |        28    29    30    31    32    33    34 | process 2 
!     5 |        35    36    37    38    39    40    41 |           
!   ----+-----------------------------------------------+-----------

!------------------------------
! Init_1dim
!
! TASK:
!   Do the one dimensional domain decomposition as depicted above.
!
! USED GLOBAL DATA:
!   OUT INTEGER start_i    : start row of the physical area this process is to calculate
!   OUT INTEGER end1_i     : end row + 1 of the physical area this process is to calculate
!   OUT INTEGER start_j    : start column of the physical area this process is to calculate
!   OUT INTEGER end1_j     : end column + 1 of the physical area this process is to calculate
!   IN/OUT INTEGER my_rank : the rank of the current process
!   OUT INTEGER comm_cart  : cartesian communicator
!   IN  INTEGER numprocs   : number of used processors
!   IN  INTEGER m          : vertical size of the physical problem
!   IN  INTEGER n          : horizontal size of the physical problem
!
! USED FUNCTIONS:
!   MPI:
!     MPI_CART_CREATE, MPI_COMM_RANK
SUBROUTINE Init_1dim()
    INTEGER :: slice_size
#ifndef serial
    INTEGER :: dims(0:1)
    LOGICAL :: periods(0:1)
    IF ((print_level >= 1) .AND. (my_rank == 0)) THEN
        WRITE(*,'(A, I3)') 'Using 1-dim domain decomposition, numprocs=', numprocs
    END IF
    ! Init the virtual topology (1-dim)
    dims(0) = numprocs
    dims(1) = 1
    periods(0) = .FALSE.
    periods(1) = .FALSE.
    CALL MPI_CART_CREATE(MPI_COMM_WORLD, 2, dims, periods, .FALSE., comm_cart, ierror)
    CALL MPI_COMM_RANK(comm_cart, my_rank, ierror)
#endif

    ! the size of each slice
    ! If m is not a multiple of numprocs, we round up the slice size
    slice_size = ((m - 1) / numprocs) + 1

    ! the rows of the physical area are distributed in slices of equal size
    ! (maybe except for the last slice) to each process
    start_i = (my_rank * slice_size)
    IF (start_i > m) THEN
        start_i = m
    END IF

    end1_i = start_i + slice_size
    IF (end1_i > m) THEN
        end1_i = m
    END IF

    ! we have the full vertical extent (all n columns) for each row 
    start_j = 0
    end1_j = n
END SUBROUTINE Init_1dim

!----------------------- 2-dim decomposition -----------------
!
!   Example: m=5, n=7, m_procs=2, n_procs=2 
! 
!       |           j=0     1     2     3  |  4     5     6            |
!   ----+----------------------------------+---------------------------+
!   i=0 |             0     1     2     3  |  4     5     6            |
!     1 | process 1   7     8     9    10  | 11    12    13  process 2 |
!     2 |            14    15    16    17  | 18    19    20            |
!   ----+----------------------------------+---------------------------+
!     3 | process 3  21    22    23    24  | 25    26    27  process 4 |
!     4 |            28    29    30    31  | 32    33    34            |
!   ----+------------------------------+-------------------------------+

!------------------------------
! Init_2dim
!
! TASK:
!   Do the two dimensional domain decomposition as depicted above.
!
! USED GLOBAL DATA:
!   OUT INTEGER start_i   : start row of the physical area this process is to calculate
!   OUT INTEGER end1_i    : end row + 1 of the physical area this process is to calculate
!   OUT INTEGER start_j   : start column of the physical area this process is to calculate
!   OUT INTEGER end1_j    : end column + 1 of the physical area this process is to calculate
!   IN  INTEGER my_rank   : the rank of the current process
!   IN  INTEGER m_procs   : vertical number of processors
!   IN  INTEGER n_procs   : horizontal number of processors
!   OUT INTEGER comm_cart : cartesian communicator
!   IN  INTEGER m         : vertical size of the physical problem
!   IN  INTEGER n         : horizontal size of the physical problem
!
! USED FUNCTIONS:
!   APPL:
!     abrt
!   MPI:
!     MPI_CART_CREATE, MPI_COMM_RANK
SUBROUTINE Init_2dim()
    INTEGER :: blocksize_i, blocksize_j
    INTEGER :: startblock_i, startblock_j
#ifndef serial
    INTEGER :: dims(0:1)
    LOGICAL :: periods(0:1)
    IF ((print_level >= 1) .AND. (my_rank == 0)) THEN
        WRITE(*,'(A, I3, A, I3, A, I3)') 'Using 2-dim domain decomposition, m_procs=', m_procs, ', &
                 & n_procs=', n_procs, ', numprocs=', numprocs
    END IF
    ! Init the virtual topology (2-dims)
    dims(0) = m_procs
    dims(1) = n_procs
    periods(0) = .FALSE.
    periods(1) = .FALSE.
    CALL MPI_CART_CREATE(MPI_COMM_WORLD, 2, dims, periods, .FALSE., comm_cart, ierror)
    CALL MPI_COMM_RANK(comm_cart, my_rank, ierror)
#endif

    IF ((my_rank == 0) .AND. ((m_procs * n_procs) /= numprocs)) THEN
        WRITE(*,*) 'Number of processors does not fit! Use ', (m_procs * n_procs), ' processors instead. &
             & Aborting...'
        CALL abrt()
    END IF

    ! the number of rows for each block (the last block might have
    ! less than the first blocks)
    blocksize_i = ((m - 1) / m_procs) + 1

    ! the number of columns for each block (the last block might have
    ! less than the first blocks)
    blocksize_j = ((n - 1) / n_procs) + 1

    ! calculate the start and end row and column
    startblock_i = my_rank / n_procs
    startblock_j = MODULO (my_rank, n_procs)
    start_i = blocksize_i * startblock_i
    end1_i = start_i + blocksize_i
    IF (end1_i > m) THEN
        end1_i = m
    END IF
    start_j = blocksize_j * startblock_j
    end1_j = start_j + blocksize_j
    IF (end1_j > n) THEN
        end1_j = n
    END IF
END SUBROUTINE Init_2dim

!------------------------------
! Init_Decomp
!
! TASK:
!   Do the domain decomposition.
!
! USED GLOBAL DATA:
!   IN/OUT INTEGER rowsize     : number of data entries in a row of the local chunk
!   IN/OUT INTEGER colsize     : number of data entries in a column of the local chunk
!   IN/OUT INTEGER chunksize   : number of data entries in the local chunk PLUS halo data
!   IN     INTEGER decomp_dims : dimension of the domain decomposition 
!   IN     INTEGER my_rank     : the rank of the current process
!
! USED FUNCTIONS:
!   APPL:
!     abrt, Init_1dim, Init_2dim
SUBROUTINE Init_Decomp()
    SELECT CASE (decomp_dims)    
        CASE (1)
            CALL Init_1dim()
        CASE (2)
            CALL Init_2dim()
        CASE DEFAULT
            WRITE(*,*) 'Wrong decomp_dims=', decomp_dims
            CALL abrt()
    END SELECT
    ! calculate the size of the rows, cols and the whole chunk
    rowsize   = end1_j - start_j
    colsize   = end1_i - start_i
    chunksize = (rowsize + 2) * (colsize + 2)

    IF ((rowsize < 2) .OR. (colsize < 2)) THEN
        WRITE(*,*) 'Each process must have at least 2 entries in each dimension.'
        WRITE(*,*) 'Decrease amount of processes or increase physical area size.'
        WRITE(*, '(A, I3, A, I0, A, I0)') 'Rank[', my_rank, ']: rowsize = ', rowsize, ' -- colsize = ', colsize
        WRITE(*,*) 'Aborting.'
        CALL abrt()
    END IF
END SUBROUTINE Init_Decomp

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 04 <<<<<<<<<<<<<<<<<<<<<<
