! CGV/solution/parts/cgvp07.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
!>>>>>>>>>>>>>>>>>>>>>> begin of task 07 <<<<<<<<<<<<<<<<<<<<<<

!************************ CG solver ****************************

!------------------------------
! CG_solve
!   OUT INTEGER num_iter : return the number of iterations
!   OUT REAL    norm_r   : return the norm of r
!
! TASK:
!   solve Laplace equation
!
! GLOBAL DATA:
!   IN     INTEGER     iter_max : maximal number of iterations
!   IN     REAL        epsilon  : epsilon for maximal error
!   IN     REAL(0:)    b        : local chunk of vector b
!   IN/OUT REAL(0:)    x        : local chunk of vector x
!   IN     INTEGER      my_rank  : rank of the local process
! USED FUNCTIONS:
!   APPL:
!     comm_vec_halo, print_vec, add_vec_vec
!     sqr_norm_vec, print_halo, mult_A_vec, dot_product_vec_vec
!   MPI-lib:
!     MPI_WTIME
SUBROUTINE CG_solve(num_iter, norm_r)
    INTEGER, INTENT(INOUT) :: num_iter
    REAL, INTENT(INOUT)    :: norm_r

    ! only p has a halo
    REAL, DIMENSION(0:chunksize-1) :: p, r, v
    REAL                           :: alpha, alpha_new, lambda
    INTEGER                        :: iter

    ! set the content of the vectors to zero
    p = 0
    r = 0
    v = 0

#ifndef serial
    ! !!! determine communication partners for vector p !!! 
    CALL Init_Halo_Struct(p)
    IF (print_level>=4) THEN
        CALL print_halo_struct()
    END IF
#endif

    ! x is initialized with 0

    ! instead of p = Ax
    ! p := x
    p = x

#ifndef serial
    ! update halo
    commtime_start = MPI_WTIME()
    CALL comm_vec_halo(p)
    commtime_end = MPI_WTIME()
    commtime_sum = commtime_sum + commtime_end - commtime_start
#endif

    ! v = Ap
    CALL mult_A_vec(p, v)
    IF (print_level>=4) THEN
        CALL print_vec('v', v)
    END IF

    ! p = b - v
    CALL add_vec_vec(b, -1.0, v, p)
    IF (print_level>=4) THEN
        CALL print_vec('p', p)
    END IF

    ! r := p
    r = p
    IF (print_level>=4) THEN
        CALL print_vec('r', r)
    END IF

    ! alpha = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2
    alpha = sqr_norm_vec(r)

    iter = 1
    DO WHILE ((iter < iter_max) .AND. (alpha > epsilon))
        iter = iter + 1
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            WRITE(*,'(A, I3, A, I0, A, F9.6)') 'CG - my_rank =', my_rank, ', i=', iter, ', alpha=', alpha
        END IF

#ifndef serial
        ! update halo
        commtime_start = MPI_WTIME()
        CALL comm_vec_halo(p)
        commtime_end = MPI_WTIME()
        commtime_sum = commtime_sum + commtime_end - commtime_start
#endif

        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('p', p)
            CALL print_halo()
        END IF

        ! v = Ap
        CALL mult_A_vec(p, v)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('v', v) 
        END IF

        ! lambda = alpha / (v, p)2 
        lambda = alpha / dot_product_vec_vec(v, p)

        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            WRITE(*,'(A, I3, A, I0, A, F9.6, A, F9.6)') 'CG - my_rank =', my_rank, ', i=', iter, ', vp=', &
                     & dot_product_vec_vec(v, p), ', lambda=', lambda
        END IF

        ! x := x + lambda * p
        CALL add_vec_vec(x, lambda, p, x)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('x', x)
        END IF

        ! r := r - lambda * v
        CALL add_vec_vec(r, -lambda, v, r)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('r', r)
        END IF

        ! alpha_new = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2 
        alpha_new = sqr_norm_vec(r)

        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            WRITE(*,'(A, I3, A, I0, A, F9.6)') 'CG - my_rank =', my_rank, ', i=', iter, ', alpha_new=', alpha_new
        END IF

        ! p := r + (alpha_new/alpha) * p 
        CALL add_vec_vec(r, alpha_new/alpha, p, p)
        IF ((print_level>=5) .OR. ( (print_level>=4) .AND. (iter==0) )) THEN
            CALL print_vec('p', p)
        END IF

        ! alpha := alpha_new
        alpha = alpha_new

        IF ((print_level>=2) .AND. (my_rank==0)) THEN
            WRITE (*,'(A, I0, A, E8.2, A)', ADVANCE='NO') 'CG i=', iter, ', ||r||**2 = ', alpha_new, ' '
            IF (alpha_new <= epsilon) THEN
                WRITE (*,'(A)', ADVANCE='NO') '<='
            ELSE
                WRITE (*,'(A)', ADVANCE='NO') '>' 
            ENDIF
            WRITE (*,'(A, E8.2, A, E8.2, A)', ADVANCE='NO') ' ', epsilon, ',   ||r|| = ', sqrt(alpha_new), ' '
            IF (SQRT(alpha_new) <= SQRT(epsilon)) THEN
                WRITE (*,'(A)', ADVANCE='NO') '<='
            ELSE
                WRITE (*,'(A)', ADVANCE='NO') '>' 
            ENDIF
            WRITE (*,'(A, E8.2)') ' ', sqrt(epsilon)
        END IF
    END DO

    ! update output-arguments
    num_iter = iter
    norm_r = SQRT(alpha)
END SUBROUTINE CG_solve

!>>>>>>>>>>>>>>>>>>>>>> -end- of task 07 <<<<<<<<<<<<<<<<<<<<<<
