! CGV/solution/parts/cgvp08.F90

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                              !
! This file has been written as a sample solution to an        !
! exercise in a course given at the High Performance           !
! Computing Centre Stuttgart (HLRS).                           !
! It is made freely available with the understanding that      !
! every copy of this file must include this header and that    !
! HLRS takes no responsibility for the use of the enclosed     !
! teaching material.                                           !
!                                                              !
! Authors: Dr. Rolf Rabenseifner                               !
!          Stephan Scheiderer                                  !
!                                                              !
! Contact: rabenseifner@hlrs.de                                !
!                                                              !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                                                            !EOH
!>>>>>>>>>>>>>>>>>>>>>> begin of task 08 <<<<<<<<<<<<<<<<<<<<<<

!************************ main program *************************

    INTEGER                         :: num_iter
    REAL                            :: norm_r, norm_err, maxnorm_err
    REAL, DIMENSION(:), ALLOCATABLE :: u, e
    REAL                            :: time_start_cpu, time_end_cpu, time_elapsed_cpu
#ifndef serial
    DOUBLE PRECISION :: time_start=0, time_end=0
    INTEGER :: dims(0:1)
    INTEGER :: percentage
#endif
    REAL :: mflops = 0, global_mflops = 0, time_elapsed, global_mflops_cpu

#ifdef serial
    numprocs = 1
    my_rank = 0
#endif
#ifndef serial
    CALL MPI_INIT(ierror)
    CALL MPI_COMM_SIZE(MPI_COMM_WORLD, numprocs, ierror)
    CALL MPI_COMM_RANK(MPI_COMM_WORLD, my_rank, ierror)
#endif

    flops = 0
    m = 4
    n = 4
    iter_max = 500
    epsilon = 1e-6
    print_level = 1
    decomp_dims = 1
    mprt = 11
    nprt = 11
    m_procs=1
    n_procs=1
#ifndef serial
    dims(0) = 0
    dims(1) = 0
    CALL MPI_DIMS_CREATE(numprocs, 2, dims, ierror)
    m_procs = dims(0)
    n_procs = dims(1)
#endif

    ! Parse command line options
    CALL read_options()

    ! Init the domain decomposition
    CALL Init_Decomp()
    IF (print_level >= 1) THEN
        CALL print_decomp()
    END IF

    ! reserve chunk of vector b and assign border data
    ALLOCATE(b(0:chunksize-1))
    b = 0
    CALL Init_Boundary_Vec(b)

    IF (print_level>=4) THEN
        CALL print_vec('b',b)
    END IF

    ! x is initialized with zeros
    ALLOCATE(x(0:chunksize-1))
    x = 0

#ifndef serial
    time_start = MPI_WTIME()
#endif
    CALL CPU_TIME(time_start_cpu)

    CALL CG_solve(num_iter, norm_r)

#ifndef serial
    time_end = MPI_WTIME()
    time_elapsed = time_end - time_start
    mflops = flops / (1000000 * time_elapsed)
    CALL MPI_REDUCE(mflops, global_mflops, 1, MPI_REAL, MPI_SUM, 0, comm_cart, ierror)
#endif
    CALL CPU_TIME(time_end_cpu)
    time_elapsed_cpu = time_end_cpu - time_start_cpu
    global_mflops_cpu = flops / (1000000 * time_elapsed_cpu)
#ifndef serial
    global_mflops_cpu = global_mflops
#endif

    ALLOCATE(u(0:chunksize-1))
    u = 0
    ALLOCATE(e(0:chunksize-1))
    e = 0
    CALL Init_Exact_Solution_vec(u)
    ! p = u - x
    CALL add_vec_vec(u, -1.0, x, e)
    norm_err = SQRT(sqr_norm_vec(e))
    maxnorm_err = max_norm_vec(e)

    IF (print_level>=3) THEN
        CALL print_vec2d('x',x,m,n)
    END IF

    IF (my_rank==0) THEN
        WRITE(*,'(A, I5, A, E8.2, A, E8.2)') 'main: ', num_iter, ' iterations, sqrt(eps) = ', SQRT(epsilon), &
                 & ',  normalized: ', SQRT(epsilon/(n*m))
        WRITE(*,'(A, E8.2, A, E8.2)') 'main:                 ||r||     = ', norm_r, ',  normalized: ', norm_r/SQRT(REAL (n*m))
        WRITE(*,'(A, E8.2, A, E8.2)') 'main:                 ||error|| = ', norm_err, ',  normalized: ', norm_err/SQRT(REAL (n*m))
        WRITE(*,'(A, E8.2, A, E8.2)') 'main:                 max(error)= ', maxnorm_err, ' = normalized: ', maxnorm_err
        WRITE(*,'(A, I0, A, I0, A, I0)') 'main:                 n*m = ', n, ' *', m, ' = ', n*m
        WRITE(*,'(A, E8.2, A)') 'main:                 ', time_elapsed_cpu, ' sec (exec. time CPU_TIME)'
#ifndef serial
        WRITE(*,'(A, E8.2, A)') 'main:                 ', time_elapsed, ' sec (exec. time MPI_WTIME)'
        WRITE(*,'(A, E8.2, A)') 'main:                 ', commtime_sum, ' sec (comm. time)'
        percentage = commtime_sum * 100.0 / time_elapsed
        WRITE(*,'(A, I2, A)') 'main: ratio (comm. time) / (exec. time): ', percentage, '%'
#endif
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (avg) CPU_TIME:  ', global_mflops_cpu / numprocs
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (all) CPU_TIME:  ', global_mflops_cpu
#ifndef serial
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (avg) MPI_WTIME:  ', global_mflops / numprocs
        WRITE(*,'(A, F8.2)') 'Calculated MFLOPS (all) MPI_WTIME:  ', global_mflops
#endif
    END IF

    IF (print_level>=1) THEN
        CALL print_vec2d('x',x,mprt,nprt)
    END IF

#ifndef serial
    CALL MPI_FINALIZE(ierror)
#endif

! beginning of the subroutine part
CONTAINS

! ********************** memory management **********************

SUBROUTINE abrt
#ifndef serial
    CALL MPI_ABORT(MPI_COMM_WORLD, -1, ierror)
#endif
    STOP 
END SUBROUTINE abrt

! ********************** commandline options ********************

SUBROUTINE read_options()
    INTEGER :: optid = 1 ! start at first optional argument 
    INTEGER :: argc
    LOGICAL :: err_detected = .FALSE.
    CHARACTER(len=255) :: arg
    argc = iargc()
    IF ( my_rank == 0 ) THEN
        DO WHILE( optid <= argc )
            CALL getarg(optid, arg) 
            IF (arg == '-m') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) m
                IF (m<=0) THEN
                    err_detected = .TRUE.
                    EXIT 
                END IF
            ELSE IF (arg == '-n') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) n
                IF (n<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                ENDIF
            ELSE IF (arg == '-imax') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) iter_max
                IF (iter_max<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                ENDIF
            ELSE IF (arg == '-eps') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) epsilon
                IF (epsilon<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE IF (arg == '-prtlev') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) print_level
                IF (print_level<0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE IF (arg == '-twodims') THEN
                decomp_dims = 2
            ELSE IF (arg == '-mprocs') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) m_procs
                IF (m_procs<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE IF (arg == '-nprocs') THEN
                optid = optid + 1
                CALL getarg(optid, arg)
                READ (arg, *) n_procs
                IF (n_procs<=0) THEN
                    err_detected = .TRUE.
                    EXIT
                END IF
            ELSE
                err_detected = .TRUE.
                EXIT
            END IF
            optid = optid + 1
        END DO
        IF (err_detected) THEN
            WRITE(*,*) 'unknown command line option: ', arg
            WRITE(*,*) 'Synopsis: mpirun -np <number of processes> ./cg_all <options>'
            WRITE(*,*) 'with options [defaults]:'
            WRITE(*,*) '  -m <number of rows in physical area of the Laplace equation> [4]'
            WRITE(*,*) '  -n <number of columns in physical area of the Laplace equation [4]'
            WRITE(*,*) '  -imax <maximal number of iterations> [500]'
            WRITE(*,*) '  -eps  <epsion> [1e-6]'
            WRITE(*,*) '  -prtlev <print and debug level 0..5> [1]'
            WRITE(*,*) '  -twodims [only one-dimensional domain decomposition (DD)]'
            WRITE(*,*) '  -mprocs <horizontal number of processes in 2-dim DD'
            WRITE(*,*) '  -nprocs <vertical   number of processes in 2-dim DD'
            WRITE(*,*) ''
            CALL abrt
        END IF
    END IF
#ifndef serial
    CALL MPI_BCAST(m,          1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(n,          1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(iter_max,   1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(epsilon,    1, MPI_REAL,       0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(print_level,1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(decomp_dims,1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(m_procs,    1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
    CALL MPI_BCAST(n_procs,    1, MPI_INTEGER,    0, MPI_COMM_WORLD, ierror);
#endif
END SUBROUTINE read_options

! >>>>>>>>>>>>>>>>>>>>>> -end- of task 08 <<<<<<<<<<<<<<<<<<<<<<
